unit TTAWriter;

(*
 * BitWriter.h
 *
 * Description: Bit writer internal interface
 *
 * Copyright (c) 2004 Alexander Djourik. All rights reserved.
 * Copyright (c) 2004 Pavel Zhilin. All rights reserved.
 *
 *)

(*
 * TTAWriter.h
 *
 * Description: TTA compressor internals
 *
 * Copyright (c) 2004 Alexander Djourik. All rights reserved.
 * Copyright (c) 2004 Pavel Zhilin. All rights reserved.
 *
 *)

(*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * aint with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Please see the file COPYING in this directory for full copyright
 * information.
 *)

 (* delphi transformation by K.Kumaki *)

interface

uses Windows, Messages, SysUtils, Classes, TTAError, TTACommon, crc32;

type
  TBitWriter = class(TObject)
  private
    hOutFile: TStream;
    start_offset: ULONG;
    bit_buffer: array [0..BIT_BUFFER_SIZE+7] of byte;
    bit_buffer_end: pbyte;
    frame_crc32: ULONG;
    bit_count: ULONG;
    bit_cache: ULONG;
    bitpos: pbyte;
    lastpos: ULONG;
    foutput_byte_count: ULONG;
  protected
  public
    constructor Create(fd: TStream; offset: ULONG);
    destructor Destroy; override;
    procedure PutHeader(var ttahdr: TTTAHeader);
    procedure PutSeekTable(seek_table: PULONG; st_size: LongInt);
    procedure PutBinary(value: ULONG; bits: ULONG);
    procedure PutUnary(value: ULONG);
    function Done: Integer;
    property output_byte_count: ULONG read foutput_byte_count;
  end;


  TTTAWriter = class(TObject)
  private
    hOutFile: TStream;
    ttahdr: TTTAHeader;
    bitWriter: TBitWriter;
    seek_table, st: PULONG;
    st_size: ULONG;
    offset, framelen, lastlen: ULONG;
    fframes, byte_size, num_chan: ULONG;
    data_pos, max_bytes: ULONG;
    tta, enc: Pencoder;
    finput_byte_count: ULONG;
    foutput_byte_count: ULONG;
  protected
  public
    constructor Create(fd: TStream; offset: LongInt; vAudioFormat: word;
        vNumChannels: word; vBitsPerSample: word;
        vSampleRate: ULONG; vDataLength: ULONG);
    destructor Destroy; override;
    function CompressBlock(data: PLongInt; data_len: LongInt): Boolean;
    property input_byte_count: ULONG read finput_byte_count;
    property output_byte_count: ULONG read foutput_byte_count;
  end;

implementation

const
  WAVE_FORMAT_PCM = 1;
  WAVE_FORMAT_IEEE_FLOAT = 3;

{ TBitWriter }

constructor TBitWriter.Create(fd: TStream; offset: ULONG);
begin
  frame_crc32 := $ffffffff;
  start_offset := offset;
  hOutFile := fd;
  bit_count := 0;
  bit_cache := 0;
  bitpos := @bit_buffer[0];
  lastpos := 0;
  foutput_byte_count := 0;
  bit_buffer_end := @bit_buffer[BIT_BUFFER_SIZE];
  hOutFile.position := start_offset;
end;

destructor TBitWriter.Destroy;
begin
  if Assigned(hOutFile) then hOutFile.position := lastpos;
  inherited Destroy;
end;

procedure TBitWriter.PutHeader(var ttahdr: TTTAHeader);
begin
  with ttahdr do begin
    TTAid := TTA1_SIGN;
    CRC32 := tta_crc32(pbyte(@ttahdr), sizeof(TTTAHeader) - sizeof(ULONG));
  end;
  if hOutFile.Write(ttahdr, sizeof(TTTAHeader)) <> sizeof(TTTAHeader) then
    raise ETTA_WRITE_ERROR.Create('BitWriter.PutHeader: Write Error');
  inc(foutput_byte_count, sizeof(TTTAHeader));
  lastpos := foutput_byte_count;
end;

procedure TBitWriter.PutSeekTable(seek_table: PULONG; st_size: LongInt);
var
  x: PULONG;
begin
  if hOutFile.position <> start_offset + sizeof(TTTAHeader) then
    hOutFile.position := start_offset + sizeof(TTTAHeader)
  else begin
    inc(foutput_byte_count, (st_size * sizeof(ULONG)));
    lastpos := foutput_byte_count;
  end;

  x := seek_table; inc(x, st_size-1);
  x^ := tta_crc32( pbyte(seek_table), (st_size - 1) * sizeof(ULONG));

  if hOutFile.Write(seek_table^, st_size * sizeof(ULONG)) <> st_size * sizeof(ULONG) then
    raise ETTA_WRITE_ERROR.Create('BitWriter.PutSeekTable: Write Error');
end;

procedure TBitWriter.PutBinary(value: ULONG; bits: ULONG);
var
 res: integer;
begin
  while bit_count >= 8 do begin
    if bitpos = bit_buffer_end then begin
      res := hOutFile.Write(bit_buffer[0], BIT_BUFFER_SIZE);
      if res <> BIT_BUFFER_SIZE then
        raise ETTA_WRITE_ERROR.Create('BitWriter.PutBinary: Write Error');
      inc(foutput_byte_count, res);
      bitpos := @bit_buffer[0];
    end;
    
    bitpos^ := byte(bit_cache and  $FF);
//  UPDATE_CRC32(bitpos^, frame_crc32);
    frame_crc32 := ((frame_crc32 shr 8) and $00FFFFFF) xor crc32_table[(frame_crc32 xor bitpos^) and $FF];
    bit_cache := bit_cache shr 8;
    dec(bit_count,8);
    inc(bitpos);
  end;
  
  bit_cache := bit_cache or ((value and bit_mask[bits]) shl bit_count);
  inc(bit_count,bits);
end;

procedure TBitWriter.PutUnary(value: ULONG);
var
 res: integer;
begin
  repeat
    while bit_count >= 8 do begin
      if bitpos = bit_buffer_end then begin
        res := hOutFile.Write(bit_buffer[0], BIT_BUFFER_SIZE);
        if res <> BIT_BUFFER_SIZE then
          raise ETTA_WRITE_ERROR.Create('BitWriter.PutUnary: Write Error');
        inc(foutput_byte_count, res);
        bitpos := @bit_buffer[0];
      end;
      
      bitpos^ := byte(bit_cache and  $FF);
//    UPDATE_CRC32(bitpos^, frame_crc32);
      frame_crc32 := ((frame_crc32 shr 8) and $00FFFFFF) xor crc32_table[(frame_crc32 xor bitpos^) and $FF];
      bit_cache := bit_cache shr 8;
      dec(bit_count,8);
      inc(bitpos);
    end;
    
    if value > 23 then begin
      bit_cache := bit_cache or (bit_mask[23] shl bit_count);
      inc(bit_count, 23);
      dec(value, 23);
    end
    else begin
      bit_cache := bit_cache or (bit_mask[value] shl bit_count);
      inc(bit_count, value + 1);
      value := 0;
    end;
  until value = 0;
end;

function TBitWriter.Done: Integer;
var
  res, bytes_to_write: ULONG;
begin
  while bit_count > 0 do begin
    bitpos^ := byte(bit_cache and $FF);
//  UPDATE_CRC32(bitpos^, frame_crc32);
    frame_crc32 := ((frame_crc32 shr 8) and $00FFFFFF) xor crc32_table[(frame_crc32 xor bitpos^) and $FF];
    bit_cache := bit_cache shr 8;
    if bit_count > 8 then dec(bit_count, 8) else bit_count := 0;
    inc(bitpos);
  end;
  
  frame_crc32 := frame_crc32 xor $FFFFFFFF;
  move(frame_crc32, bitpos^, 4);
  bytes_to_write := ULONG(bitpos) + sizeof(ULONG) - ULONG(@bit_buffer[0]);
  
  res := hOutFile.Write(bit_buffer[0], bytes_to_write);
  if res <> bytes_to_write then
    raise ETTA_WRITE_ERROR.Create('BitWriter.Done: Write Error');
  
  inc(foutput_byte_count, res);
  bitpos := @bit_buffer[0];
  frame_crc32 := $FFFFFFFF;
  
  Result := foutput_byte_count - lastpos;
  lastpos := foutput_byte_count;
end;

{ TTTAWriter }

constructor TTTAWriter.Create(fd: TStream; offset: LongInt; vAudioFormat: word;
        vNumChannels: word; vBitsPerSample: word;
        vSampleRate: ULONG; vDataLength: ULONG);
begin
  hOutFile := fd;
  with ttahdr do begin
    AudioFormat := vAudioFormat;
    NumChannels := vNumChannels;
    BitsPerSample := vBitsPerSample;
    SampleRate := vSampleRate;
    DataLength := vDataLength;
  end;
  
  if ttahdr.AudioFormat <> WAVE_FORMAT_PCM then
    raise ETTA_FORMAT_ERROR.Create('TTAWriter.Create: Format Error');
  if ttahdr.BitsPerSample = MAX_BPS then
    raise ETTA_FORMAT_ERROR.Create('TTAWriter.Create: Format Error');
  
  framelen := Trunc(FRAME_TIME * ttahdr.SampleRate+1e-9);
  num_chan := ttahdr.NumChannels; // shl is_float;
  byte_size := (ttahdr.BitsPerSample + 7) div 8;
  
  finput_byte_count := 0;
  foutput_byte_count := 0;
  
  fframes := ttahdr.DataLength div framelen;
  lastlen := ttahdr.DataLength - fframes * framelen;
  if lastlen > 0 then inc(fframes);
  st_size := (fframes + 1);
  
  // grab some space for an encoder buffers
  getmem(seek_table, sizeof(ULONG) * st_size);
  st := seek_table;
  getmem(tta, sizeof(Tencoder) * num_chan);
  enc := tta;
  bitWriter := TBitWriter.Create(hOutFile, offset);
  bitWriter.PutHeader(ttahdr);
  bitWriter.PutSeekTable(seek_table, st_size);
  
  data_pos := 0;
  tta_encoder_init(tta, num_chan, byte_size);
  dec(fframes);
  
  max_bytes := (num_chan * byte_size * ttahdr.DataLength); // shr is_float;
end;

destructor TTTAWriter.Destroy;
begin
  if Assigned(bitWriter) and Assigned(seek_table) then
  begin
    bitWriter.PutSeekTable(seek_table, st_size);
  end;
  if Assigned(bitWriter) then bitWriter.Free;
  if Assigned(seek_table) then Freemem(seek_table);
  if Assigned(tta) then Freemem(tta);
  inherited Destroy;
end;

function TTTAWriter.CompressBlock(data: PLongInt; data_len: Longint): Boolean;
var
  p, q, r, s: PLongint;
  tmp, prev: Longint;
  value, k, i, unary, binary: ULONG;
  len: LongInt;
  fst: Pfltst;
  rice: Padapt;
  last: PLongInt;
  pA, pB, pM: PLongInt;
  sum: LongInt;
  b: byte;
begin
  Result := False;
  
  if data_len > (max_bytes - finput_byte_count) then
    data_len := (max_bytes - finput_byte_count);
  
  finput_byte_count := finput_byte_count +
                      (num_chan * byte_size * data_len); // shr is_float;
  
  while (data_len > 0) and (fframes >= 0) do begin
    if data_len < LongInt(framelen - data_pos) then
      len := data_len
    else
      len := framelen - data_pos;
    dec(data_len, len);
    
    p := data; s := p; inc(s);
    prev := 0;

    for i := 0 to len * num_chan-1 do begin
      fst := @(enc.fst);
      rice := @(enc.rice);
      last := @(enc.last);

      // transform data
      if ULONG(enc) < ULONG(tta) + (num_chan - 1) * sizeof(Tencoder) then begin
        prev := s^ - p^;
        p^ := prev;
      end
      else p^ := p^ - prev div 2;

      tmp := p^;
      case byte_size of
       1: p^ := p^ - LongInt(((uint64(last^) shl 4) - last^) shr 4); // bps 8
       2: p^ := p^ - LongInt(((uint64(last^) shl 5) - last^) shr 5); // bps 16
       3: p^ := p^ - LongInt(((uint64(last^) shl 5) - last^) shr 5); // bps 24
       4: p^ := p^ - last^; // bps 32
      end;
      last^ := tmp;
      
      // compress stage 2: adaptive hybrid filter
//    hybrid_filter(fst, p, 1);
      pA := @fst.dl[0];
      pB := @fst.qm[0];
      pM := @fst.dx[0];
      sum := fst.round;
      
      if fst.error = 0 then begin
        inc(sum, pA^ * pB^); inc(pA); inc(pB);
        inc(sum, pA^ * pB^); inc(pA); inc(pB);
        inc(sum, pA^ * pB^); inc(pA); inc(pB);
        inc(sum, pA^ * pB^); inc(pA); inc(pB);
        inc(sum, pA^ * pB^); inc(pA); inc(pB);
        inc(sum, pA^ * pB^); inc(pA); inc(pB);
        inc(sum, pA^ * pB^); inc(pA); inc(pB);
        inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM, 8);
      end
      else if fst.error < 0 then begin
        dec(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        dec(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        dec(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        dec(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        dec(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        dec(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        dec(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        dec(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
      end
      else begin
        inc(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        inc(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        inc(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        inc(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        inc(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        inc(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        inc(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
        inc(pB^, pM^); inc(sum, pA^ * pB^); inc(pA); inc(pB); inc(pM);
      end;

    q := pM; r := pA; dec(r);
    value := r^;
    asm
      SAR     value,30
      OR      value,1
      SAL     value,2
    end;
    q^ := value;
    dec(q); dec(r);
    value := r^;
    asm
      SAR     value,30
      OR      value,1
      SAL     value,1
    end;
    q^ := value;
    dec(q); dec(r);
    value := r^;
    asm
      SAR     value,30
      OR      value,1
      SAL     value,1
    end;
    q^ := value;
    dec(q); dec(r);
    value := r^;
    asm
      SAR     value,30
      OR      value,1
    end;
    q^ := value;

//  if mode <> 0 then begin
      pA^ := p^;
      b := fst.shift;
      asm
        MOV CL,b
        SAR sum,CL
      end;
      dec(p^, sum);
      fst.error := p^;
//  end

      if fst.mutex <> 0 then begin
        q := pA; r := q; dec(r);
        r^ := q^ - r^; dec(q); dec(r);
        r^ := q^ - r^; dec(q); dec(r);
        r^ := q^ - r^;
      end;

//    memshl (fs->dl, fs->dl + 1);
      q := @fst.dl[0]; r := q; inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^;

//    memshl (fs->dx, fs->dx + 1);
      q := @fst.dx[0]; r := q; inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^; inc(q); inc(r);
      q^ := r^;
// end of hybrid_filter

//    value := ENC(p^);
      if p^ > 0 then value := (p^ shl 1) - 1 else value := - p^ shl 1;
      
      // encode Rice unsigned
      k := rice.k0;
      
      rice.sum0 := rice.sum0 + value - (rice.sum0 shr 4);
//    if (rice.k0 > 0) and (rice.sum0 < shift_16[rice.k0]) then
      if (k > 0) and (rice.sum0 < shift_16[k]) then
        dec(rice.k0)
//    else if rice.sum0 > shift_16[rice.k0 + 1] then
      else if rice.sum0 > shift_16[k + 1] then
        inc(rice.k0);
      
      if value >= bit_shift[k] then begin
        dec(value, bit_shift[k]);
        k := rice.k1;
        
        rice.sum1 := rice.sum1 + value - (rice.sum1 shr 4);
//      if (rice.k1 > 0) and (rice.sum1 < shift_16[rice.k1]) then
        if (k > 0) and (rice.sum1 < shift_16[k]) then
          dec(rice.k1)
//      else if rice.sum1 > shift_16[rice.k1 + 1] then
        else if rice.sum1 > shift_16[k + 1] then
          inc(rice.k1);
        
        unary := 1 + (value shr k);
      end
      else unary := 0;
      
      bitWriter.PutUnary(unary);
      if k <> 0 then begin
        binary := value and bit_mask[k];
        bitWriter.PutBinary(binary, k);
      end;
      
      if ULONG(enc) < ULONG(tta) + (num_chan - 1) * sizeof(Tencoder) then inc(enc)
      else enc := tta;
      
      inc(p); inc(s);
    end;
    
    inc(data_pos, len);
    
    if data_pos = framelen then begin
      st^ := bitWriter.Done; inc(st);
      
      dec(fframes);
      if (fframes = 0) and (lastlen <> 0) then framelen := lastlen;
      tta_encoder_init(tta, num_chan, byte_size);
      data_pos := 0;
    end;
    
    inc(data, (num_chan * len));
    
    Result := True;
  end;
  
  foutput_byte_count := bitWriter.output_byte_count;
end;

end.
